// N-body simulation: sequential implementation

#include <stdio.h>
#include <math.h>
#define INCLUDE_PARTICLE_COMPUTATION_FUNCS
#include "sim.h"

// Struct type representing the overall simulation
// (note that this is given the typedef name "Simulation"
// in sim.h)
struct Simulation_ {
	struct drand48_data rng;
	int num_particles;
	ParticleData pd;
};

Simulation *sim_create(SimulationParams *params)
{
	Simulation *sim = (Simulation *) malloc(sizeof(Simulation));

	// Seed the random number generator
	srand48_r(SEED, &sim->rng);

	sim->num_particles = params->num_particles;

	// Create and initialize host ParticleData
	particle_data_alloc_buffers(&sim->pd, sim->num_particles);
	for (int i = 0; i < sim->num_particles; i++) {
		particle_init_rand(sim->pd.x, sim->pd.y, sim->pd.dx, sim->pd.dy, sim->pd.mass, sim->pd.color, i, &sim->rng);
	}

	return sim;
}

int sim_get_num_particles(Simulation *sim)
{
	return sim->num_particles;
}

float sim_get_particle_x(Simulation *sim, int index)
{
	return sim->pd.x[index];
}

float sim_get_particle_y(Simulation *sim, int index)
{
	return sim->pd.y[index];
}

int sim_get_particle_color(Simulation *sim, int index)
{
	return sim->pd.color[index];
}

void sim_destroy(Simulation *sim)
{
	particle_data_free_buffers(&sim->pd);
	free(sim);
}

void sim_tick(Simulation *sim)
{
	// Simulate the force on each particle due to the gravitational attraction
	// to all of other particles, and update each particle's velocity accordingly.
	for (int i = 0; i < sim->num_particles; i++) {
		for (int j = 0; j < sim->num_particles; j++) {
			if (i != j) {
				particle_compute_attraction(sim->pd.x, sim->pd.y, sim->pd.dx, sim->pd.dy, sim->pd.mass, i, j);
			}
		}
	}

	// Based on each particle's velocity, update its position.
	for (int i = 0; i < sim->num_particles; i++) {
		particle_update_position(sim->pd.x, sim->pd.y, sim->pd.dx, sim->pd.dy, i);
	}
}
